/*
 * OCSP responder
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA project 2001
 *
 * Copyright (c) 2001 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "general.h"
#include "http_client.h"

#include <openssl/txt_db.h>
#include <openssl/conf.h>
#include <openssl/lhash.h>
#include <openssl/asn1.h>

#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/bn.h>
#include <openssl/pem.h>
#include <openssl/bio.h>

#ifdef HAVE_OPENLDAP
#include <ldap.h>
#endif

#include "support.h"
#include "ocsp_db.h"

/* External imported variables */
extern int debug;
extern int verbose;

/* Functions */
X509_CRL *ocspd_get_crl ( URL *url ) {
#ifdef _DEBUG
	BIO *debug = NULL;
#endif
	X509_CRL *crl = NULL;

	if( url->proto == OCSP_CRL_PROTO_LDAP ) {
#ifdef HAVE_OPENLDAP
		crl = ocspd_load_crl_ldap(url);
#else
		syslog(LOG_ERR, 
			"Error, LDAP support was not enabled when compiling.");
		return NULL;
#endif
	} else if (  url->proto == OCSP_CRL_PROTO_FILE ) {
		crl = ocspd_load_crl_file(url);
	} else if (  url->proto == OCSP_CRL_PROTO_HTTP ) {
		crl = ocspd_load_crl_http(url);
	} else {
		return NULL;
	}

#ifdef _DEBUG
	if(verbose && crl) {
		debug=BIO_new_fp(stdout,BIO_NOCLOSE);
		X509_CRL_print( debug, crl );
		BIO_free(debug);
	}
#endif

	return crl;

}

STACK_OF(X509) *ocspd_get_ca ( URL *url ) {

	STACK_OF(X509) *ret = NULL;

	if( url->proto == OCSP_CRL_PROTO_LDAP ) {
#ifdef HAVE_OPENLDAP
		ret = ocspd_load_ca_ldap(url);
#else
		syslog(LOG_ERR, 
			"Error, LDAP support was not enabled when compiling.");
		return NULL;
#endif
	} else if (  url->proto == OCSP_CRL_PROTO_FILE ) {
		ret = ocspd_load_ca_file(url);
	} else if (  url->proto == OCSP_CRL_PROTO_HTTP ) {
		ret = ocspd_load_ca_http(url);
	} else {
		return NULL;
	}

	return ret;
}

#ifdef HAVE_OPENLDAP

LDAP *ocspd_ldap_connect( URL *url ) {

	LDAP	*ld = NULL;
	int	protocol = -1;
	int	ret = 0;

	
        (void) signal( SIGPIPE, SIG_IGN );

	if( (!url) || (!url->addr)) {
		syslog( LOG_ERR, "ERROR, no LDAP server address set!");
		return NULL;
	}

	if( !(url->dn) ) {
		syslog( LOG_ERR, "ERROR: no DN set for %s!", url->addr );
		return NULL;
	}

	if(verbose)
		syslog( LOG_INFO, "Connecting to LDAP (%s)", url->addr );

	if (( ld = ldap_init( url->addr, url->port )) == NULL ) {
		syslog( LOG_ERR, "ldap_init failure!");
		return NULL;
	}

	if(verbose)
		syslog( LOG_INFO, "Connection established (%s)",
			url->addr );

	/* Setting LDAP protocol V3 */
	protocol = LDAP_VERSION3;
	if( ldap_set_option( ld, LDAP_OPT_PROTOCOL_VERSION, &protocol )
						!= LDAP_OPT_SUCCESS ) {
		syslog( LOG_ERR, "Could not set LDAP_OPT_PROTOCOL_VERSION %d\n",
                                 protocol );
		ldap_unbind(ld);
		return(NULL);
	}

	if ( (ret = ldap_bind_s( ld, url->usr, url->pwd, LDAP_AUTH_SIMPLE )) 
			!= LDAP_SUCCESS ) {
		syslog( LOG_ERR, "ldap_bind_s failure (%s:%d) [%d]",
		     url->addr, url->port, ret );
		// ldap_perror( ld, "ldap_bind" );
		return NULL;
	}

	if( verbose )
		syslog( LOG_INFO, "Successfully binded to LDAP (%s)",
			url->dn);

	return(ld);
}

STACK_OF(X509) *ocspd_load_ca_ldap( URL *url ) {

	LDAP	*ld = NULL;
	int  	i,rc;
	char	*attrs[] = { url->attr, NULL };
	char    *filter = "objectclass=*";
	struct berval **vals = NULL;
	LDAPMessage *res = NULL;
	BIO	*membio = NULL;

	X509 *x = NULL;
	STACK_OF(X509) * x_sk = NULL;

	if( (url->attr == NULL) || (url->dn == NULL)) {
		return NULL;
	}

	if((ld = ocspd_ldap_connect ( url )) == NULL ) {
		return NULL;
	}

	/* We search for the exact match, so LDAP_SCOPE_BASE is used here */
	if (( rc = ldap_search_s( ld, url->dn, LDAP_SCOPE_BASE,
			filter, attrs, 0, &res )) != LDAP_SUCCESS ) {

		if ( rc != LDAP_NO_SUCH_OBJECT ) {
			syslog( LOG_ERR, 
				"LDAP: [%s] object not found", url->dn );
		}

		goto end;
	} else {
		if(verbose)
			syslog( LOG_INFO, "LDAP: Search Successful");
	}

	i = ldap_count_entries( ld, res );
	if( verbose )
		syslog( LOG_INFO, "Got %d LDAP entries", i );

	if((vals = ldap_get_values_len (ld, res, attrs[0])) != NULL ) {

		if((x_sk = sk_X509_new_null()) == NULL ) {
			return NULL;
		}

		for( i=0; vals[i] != NULL; i++ ) {
			/*
			int c;
			char buf[256];
			*/
		
			/* create a new memory bio where pushing the X509
			 * data and then load using the ocspd_X509_bio
			 * function */
			membio = BIO_new_mem_buf( vals[i]->bv_val, 
							vals[i]->bv_len );

			if(!membio) break;

		        /* Try PEM format */
		        if((x=PEM_read_bio_X509(membio,NULL,NULL,NULL)) 
								== NULL ) {

				BIO_reset( membio );
				ERR_clear_error();

                		/* Is it DER encoded (???) */
                		if((x=d2i_X509_bio(membio,NULL)) == NULL ) {
                        		syslog(LOG_ERR,
						"LDAP: ERROR CA Cert format UNKNOWN!");
                		} else {
                        		if(verbose)
                                		syslog(LOG_INFO,
                                        	"LDAP: CA Cert is DER formatted");
                		}
        		} else {
                		if(verbose)
                		        syslog(LOG_INFO,
						"LDAP: CA Cert is PEM formatted");
        		}


			if( x ) {
				if(verbose)
					syslog( LOG_INFO, 
						"LDAP: Got CA cert from LDAP");
				sk_X509_push( x_sk, x );
			} else {
				syslog( LOG_ERR, 
					"LDAP:ERROR can not read CA Cert");
			}

			BIO_free( membio );
		}
		ldap_value_free_len( vals );
		vals = NULL;
	} else {
		syslog( LOG_ERR, "LDAP: No CA Cert data found!" );
		return NULL;
	}

end:
	if(res) ldap_msgfree( res );
	if(ld) ldap_unbind( ld );

	if( verbose )
		syslog( LOG_INFO, "LDAP: Successfully unbinded");

	return x_sk;
}

X509_CRL *ocspd_load_crl_ldap( URL *url ) {
	LDAP	*ld = NULL;
	int  	i = 0;
	int	rc = 0;
	int	ret = 0;
	char	*attrs[] = { url->attr, NULL };
	char    *filter = "objectclass=*";
	struct berval **vals = NULL;
	LDAPMessage *res = NULL;
	BIO	*membio = NULL;

	X509_CRL *crl = NULL;

	if((ld = ocspd_ldap_connect ( url )) == NULL ) {
		return NULL;
	}

	if ( (ret = ldap_bind_s( ld, url->usr, url->pwd, LDAP_AUTH_SIMPLE )) 
			!= LDAP_SUCCESS ) {
		syslog( LOG_ERR, "LDAP: ldap_bind_s failure (%s:%d) [%d]",
		     url->addr, url->port, ret );
		// ldap_perror( ld, "ldap_bind" );
		return NULL;
	}

	if( verbose )
		syslog( LOG_INFO, "LDAP: Successfully binded (%s)", url->dn);

	// if (( rc = ldap_search_s( ld, url->dn, LDAP_SCOPE_SUBTREE,

	/* We search for the exact match, so LDAP_SCOPE_BASE is used here */
	if (( rc = ldap_search_s( ld, url->dn, LDAP_SCOPE_BASE,
			filter, attrs, 0, &res )) != LDAP_SUCCESS ) {

		if ( rc != LDAP_NO_SUCH_OBJECT ) 
			syslog( LOG_ERR, "LDAP ERROR: [%s] NOT found", 
				url->dn );

		goto end;
	} else {
		if(verbose)
			syslog( LOG_INFO, "LDAP: Search Successful");
	}

	i = ldap_count_entries( ld, res );
	if( verbose )
		syslog( LOG_INFO, "LDAP: Got [%d] entries", i );

	if((vals = ldap_get_values_len(ld, res, attrs[0])) != NULL ) {

		for( i=0; vals[i] != NULL; i++ ) {
			/* create a new memory bio where pushing the CRL
			 * data and then load using the ocspd_X509_CRL_bio
			 * function */
			membio = BIO_new_mem_buf( vals[0]->bv_val, 
							vals[0]->bv_len );

			/* Load CRL from the membio */
			if( (crl = ocspd_X509_CRL_bio( membio, FORMAT_ASN1 ))
						== NULL ) {
				BIO_reset( membio );
				ERR_clear_error();
				crl = ocspd_X509_CRL_bio( membio, FORMAT_PEM );
			}

			if( crl ) {
				if(verbose)
					syslog( LOG_INFO, "LDAP: Got CRL ok.");
			} else {
				syslog( LOG_ERR, "LDAP: can not read CRL data");
			}

			BIO_free( membio );
		}
		ldap_value_free_len( vals );
	} else {
		syslog( LOG_ERR, "LDAP: No CRL data found!");
		return NULL;
	}

end:
	if(res) ldap_msgfree( res );
	if(ld) ldap_unbind( ld );

	if( verbose )
		syslog( LOG_INFO, "LDAP: Successfully unbinded" );

	return(crl);
}
#endif


STACK_OF(X509) *ocspd_load_ca_http( URL *url ) {
	int cont = 1;
	int head_lines = 0;
	char linebuf[1024];
	X509 *ret = NULL;
	STACK_OF(X509) *x_sk = NULL;
	BIO *mem = NULL;
	BUF_MEM *buf = NULL;

	/* Check for non empty URL */
	if( (!url) || (!url->addr)) {
		syslog(LOG_ERR, "Configuration error: needed valid URL for"
			" CA Cert HTTP download!");
		return NULL;
	}

	/* Get the object data from HTTP */
	if( (buf = ocspd_http_get ( url, 0 )) == NULL ) {
		syslog(LOG_ERR, "Error loading CRL from HTTP");
		syslog(LOG_ERR, "Error loading CA Cert from HTTP [%s]",
			url->addr );
		return NULL;
	}

	if( verbose ) {
		syslog( LOG_INFO, "Successfully got CA HTTP data [%lu]",
			buf->length);
	}

	/* Build a memory bio from the MEM_BUF data */
	if(!(mem = BIO_new_mem_buf(buf->data, buf->length) )) {
		syslog( LOG_ERR, "ERROR: Internal memory allocation error!");
		if( buf ) BUF_MEM_free (buf);
		return NULL;
	}

	if ( ocspd_parse_http_headers ( mem ) == 0 ) {
		if( mem ) BIO_free_all(mem);
		if( buf ) BUF_MEM_free ( buf );
		return NULL;
	}

	if((x_sk = sk_X509_new_null()) == NULL ) {
		syslog(LOG_ERR, "ERROR allocating memory.");
		if( mem ) BIO_free_all(mem);
		if( buf ) BUF_MEM_free ( buf );
		return NULL;
	}

	/* Try PEM format */
	if ( (ret=PEM_read_bio_X509(mem,NULL,NULL,NULL)) == NULL ) {
		BIO_reset(mem);
		ERR_clear_error();
		if ( ocspd_parse_http_headers ( mem ) == 0 ) {
			if( mem ) BIO_free_all(mem);
			if( buf ) BUF_MEM_free ( buf );
			return NULL;
		}

		/* Is it DER encoded (???) */
		if( (ret = d2i_X509_bio( mem, NULL )) == NULL ) {
			syslog( LOG_ERR, "ERROR -- Unknown CA cert"
				" format (should either PEM or DER)");
			ERR_clear_error();
		} else {
			if( verbose )
				syslog(LOG_INFO, "CA cert is in DER format");
		}
	} else {
		if( verbose )
			syslog(LOG_INFO, "CA cert is in PEM format");
	}

	if( ret )
		sk_X509_push(x_sk, ret);

	if( mem ) BIO_free_all(mem);
	if( buf ) BUF_MEM_free ( buf );

	if(verbose) {
		syslog(LOG_INFO, "Loaded %d CA Certificates", 
						sk_X509_num(x_sk));
	}

	return x_sk;
}

X509_CRL *ocspd_load_crl_http( URL *url ) {
	int cont = 1;
	X509_CRL *crl = NULL;
	BIO *mem = NULL;
	BUF_MEM *buf = NULL;

	int head_lines = 0;
	char linebuf[1024];

	/* Check for non empty URL */
	if( (!url) || (!url->addr)) {
		syslog(LOG_ERR, "Configuration error: needed valid URL for"
			" CA Cert HTTP download!");
		return NULL;
	}

	/* Get the object data from HTTP */
	if( (buf = ocspd_http_get ( url, 0 )) == NULL ) {
		syslog(LOG_ERR, "Error loading CRL from HTTP");
		syslog(LOG_ERR, "Error loading CRL from HTTP [%s]",
			url->addr );
		return NULL;
	}

	if( verbose ) {
		syslog( LOG_INFO, "Successfully got CRL HTTP data [%lu]",
			buf->length);
	}

	/* Build a memory bio from the MEM_BUF data */
	if(!(mem = BIO_new_mem_buf(buf->data, buf->length) )) {
		syslog( LOG_ERR, "ERROR: Internal memory allocation error!");
		if( buf ) BUF_MEM_free (buf);
		return NULL;
	}

	/* Check the HTTP headers */
	if( ocspd_parse_http_headers( mem ) == 0 ) {
		if( mem ) BIO_free_all (mem);
		if( buf ) BUF_MEM_free (buf);
		return NULL;
	}

	/* Try and load CRL - ASN1 first, PEM second */
	if( (crl = ocspd_X509_CRL_bio( mem, FORMAT_ASN1 )) == NULL ) {

		/* If it is not in DER format, let's try the PEM one */
		BIO_reset(mem);
		ERR_clear_error();

		/* Parse the headers again... */
		if( ocspd_parse_http_headers( mem ) == 0 ) {
			if( mem ) BIO_free_all (mem);
			if( buf ) BUF_MEM_free (buf);
			return NULL;
		}

                if( (crl = ocspd_X509_CRL_bio( mem, FORMAT_PEM )) == NULL ) {
			syslog(LOG_ERR, "ERROR -- unknown CRL format (should "
				"be PEM or DER encoded)");
			ERR_clear_error();
		} else {
			if(verbose)
				syslog(LOG_INFO, "CRL is in PEM format");
		}
        } else {
		if (verbose)
			syslog(LOG_INFO, "CRL is in DER format");
	}

	if( mem ) BIO_free(mem);
	if( buf ) BUF_MEM_free( buf);

	return crl;
}

STACK_OF(X509) *ocspd_load_ca_file( URL *url ) {

	int cont = 1;
	X509 *ret = NULL;
	BIO *in = NULL;
	STACK_OF(X509) *x_sk = NULL;

	if ((in=BIO_new_file( url->addr, "r")) == NULL) {
		syslog(LOG_ERR,"Can not open file %s.", url->addr );
		return(NULL);
	}

	cont = 1;
	if((x_sk = sk_X509_new_null()) == NULL ) {
		return(NULL);
	}

	while ( cont == 1 ) {
		/* Try PEM format */
		if ( (ret =PEM_read_bio_X509(in,NULL,NULL,NULL)) == NULL ) {
			/* Is it DER encoded (???) */
			if((ret=d2i_X509_bio(in,NULL)) == NULL ) {
				// syslog(LOG_ERR,"CA Cert UNKNOWN [%s]!",
				// 		url->addr );
				cont = 0;
				break;
			}
			/* else {
				if(verbose)
					syslog(LOG_INFO,
						"CA Cert is DER formatted.");
			}
			*/
		}
		/* else {
			if(verbose)
				syslog(LOG_INFO,"CA Cert is PEM formatted.");
		}
		*/

		if(!ret) {
			cont = 0;
		} else {
			sk_X509_push(x_sk, ret);
		}
	}

	return x_sk;
}

X509_CRL *ocspd_load_crl_file( URL *url ) {
	BIO *in       = NULL;
	X509_CRL *crl = NULL;

	/* Load CRL */
	if ((in=BIO_new_file( url->addr, "r")) == NULL) {
		return(NULL);
	}

	if( (crl = ocspd_X509_CRL_bio( in, FORMAT_PEM )) == NULL ) {
		if((crl = ocspd_X509_CRL_bio( in, FORMAT_ASN1 )) == NULL ) {
			syslog(LOG_ERR, "ERROR -- unknown CRL format (should "
				"be PEM or DER encoded)");
		} else {
			if(verbose)
				syslog(LOG_INFO, "CRL is in DER format");
		}
	} else {
		if(verbose)
			syslog(LOG_INFO, "CRL is in PEM format");
	}

	BIO_free( in );

	/*
	rev = X509_CRL_get_REVOKED(crl);

	for(i = 0; i < sk_X509_REVOKED_num(rev); i++) {
		r = sk_X509_REVOKED_value(rev, i);
		BIO_printf(out,"    Serial Number: ");
		i2a_ASN1_INTEGER(out,r->serialNumber);
		BIO_printf(out,"\n        Revocation Date: ","");
		ASN1_TIME_print(out,r->revocationDate);
		BIO_printf(out,"\n");
		X509V3_extensions_print(out, "CRL entry extensions",
		r->extensions, 0, 8);
	}
	*/

	return crl;
}

X509_CRL *ocspd_X509_CRL_bio ( BIO *in, int form ) {

	X509_CRL *crl = NULL;

	if (form == FORMAT_ASN1)
		/* Format is DER encoded */
		crl=d2i_X509_CRL_bio(in,NULL);
	else if (form == FORMAT_PEM)
		/* Format is PEM encoded */
		crl=PEM_read_bio_X509_CRL(in,NULL,NULL,NULL);
	else	{
		return(NULL);
		}

	return crl;
}

int ocspd_verify_cert ( X509_CRL *crl, ASN1_INTEGER *serial ) {

	int ok;
	X509_REVOKED rtmp;

	/* Look for serial number of certificate in CRL */
	rtmp.serialNumber = serial;
	ok = sk_X509_REVOKED_find(crl->crl->revoked, &rtmp);

	/* Not found: OK */
	if(ok == -1) return 1;

	/* Returns 0 if it does not find the certificate within the
	 * provided CRL */
	return 0;
}

size_t my_min ( size_t a, size_t b ) {
	if (a > b) {
		return(b);
	} else {
		return(a);
	}
}

STACK_OF(CA_ENTRY_CERTID) *ocspd_CA_ENTRY_CERTID_new ( 
				STACK_OF(X509)* ca_cert, EVP_MD *digest ) {

	CA_ENTRY_CERTID *cid = NULL;
	X509_NAME *iname = NULL;
	ASN1_BIT_STRING *ikey = NULL;
#ifdef _DEBUG
	BIO *bio = NULL;
#endif
	STACK_OF(CA_ENTRY_CERTID) *cid_sk = NULL;
	X509 *x = NULL;

	unsigned char buff[EVP_MAX_MD_SIZE];

	unsigned int j, i, nid;

	/* Check for needed info */
	if( (ca_cert == NULL ) || (digest == NULL)) 
		return (NULL);

	if((cid_sk = sk_CA_ENTRY_CERTID_new_null()) == NULL ) {
		return (NULL);
	}

	if((cid = OPENSSL_malloc ( sizeof(CA_ENTRY_CERTID) ))== NULL ) {
		return cid_sk;
	}

	for ( j = 0; j < sk_X509_num( ca_cert ); j++ ) {
		x = sk_X509_value( ca_cert, j );
		ikey = X509_get0_pubkey_bitstr( x );
		iname = X509_get_subject_name ( x );
#ifdef _DEBUG
		/*
		bio = BIO_new ( BIO_s_file());
		BIO_set_fp( bio, stderr, BIO_NOCLOSE);

		fprintf( stderr, "CA nameHash =>");
		*/
#endif
		X509_NAME_digest( iname, digest, buff, &i );
		if( !(cid->nameHash = ASN1_OCTET_STRING_new()) ||
			   !(ASN1_OCTET_STRING_set(cid->nameHash, buff, i))) {
			if( cid ) OPENSSL_free (cid);
			return (NULL);
		};

		EVP_Digest( ikey->data, ikey->length, buff, &i, digest, NULL);
		if( !(cid->keyHash = ASN1_OCTET_STRING_new()) ||
				!(ASN1_OCTET_STRING_set(cid->keyHash,buff,i))) {
			if( cid ) OPENSSL_free (cid);
			return (NULL);
		};

		/* Set the Digest Algorithm used */
		nid = EVP_MD_type(digest);
		if (nid == NID_undef ) {
			if( cid ) OPENSSL_free (cid);
			return (NULL);
		};
		cid->hashAlgorithm = (X509_ALGOR *) OBJ_nid2obj(nid);
		if(cid->hashAlgorithm == NULL) {
			if( cid ) OPENSSL_free (cid);
			return (NULL);
		};
		cid->hashAlgorithm->parameter = ASN1_TYPE_new();
		if(cid->hashAlgorithm->parameter == NULL ) {
			if( cid->hashAlgorithm ) X509_ALGOR_free( cid->hashAlgorithm);
			if( cid ) OPENSSL_free (cid);
			return (NULL);
		}
		cid->hashAlgorithm->parameter->type = V_ASN1_NULL;
		sk_CA_ENTRY_CERTID_push( cid_sk, cid );
	}

	return( cid_sk );
}

STACK_OF(X509_REVOKED) *ocspd_build_crl_entries_list ( CA_LIST_ENTRY *ca,
			X509_CRL *crl ) {

	long rev_num = 0;
	/*
	long def_status = OCSP_REVOKED_STATUS_UNSPECIFIED;
	int i;
	*/

	/*
	X509_CRL_ENTRY *entry = NULL;
	X509_REVOKED *r = NULL;
	*/

	STACK_OF(X509_REVOKED) *ret = NULL;

	if( !ca || !crl ) return (NULL);

	ret = X509_CRL_get_REVOKED(crl);
	rev_num = sk_X509_REVOKED_num(ret);

	if( verbose )
		syslog( LOG_INFO, "%ld CRL Entries [ %s ]", 
			rev_num, ca->ca_id );

	ca->crl_list = ret;
	ca->entries_num = rev_num;

	if( (rev_num > -1 ) && (ca->crl_list == NULL) ) {
		syslog( LOG_ERR, "%s:%d Memory allocation error!",
			__FILE__, __LINE__ );
		return NULL;
	}

	sk_X509_REVOKED_sort(ca->crl_list);

	return (ca->crl_list);
}

CA_LIST_ENTRY *ocspd_CA_ENTRY_find ( OCSPD_CONFIG *conf, OCSP_CERTID *cid ) {
	long i, j;
	int ret;

	OCSP_CERTID *b = NULL;
	STACK_OF(CA_ENTRY_CERTID) *a = NULL;
	CA_ENTRY_CERTID *tmp = NULL;

	b = cid;

	if( (conf == NULL) || ( conf->ca_list == NULL ))
		return NULL;

	for( i = 0; i < conf->ca_list_len; i++ ) {

		a = conf->ca_list[i]->cid;

		/*
		if( verbose )
			syslog( LOG_INFO, "%s:%d Debug", __FILE__, __LINE__);
		*/

		/* Check the Hash Algorithm */
		/*
		ret = OBJ_cmp(a->hashAlgorithm->algorithm, 
					b->hashAlgorithm->algorithm);
		if( verbose )
			syslog( LOG_INFO, "%s:%d Debug ret=%d", 
				__FILE__, __LINE__, ret);

		if (ret) {
			syslog( LOG_ERR, "%s:%d CA mismatch (%d)"
				__FILE__,__LINE__, ret);
			continue;
		}
		*/

		for( j = 0; j < sk_CA_ENTRY_CERTID_num(a); j++ ) {

			if((tmp = sk_CA_ENTRY_CERTID_value(a,j))==NULL )
				break;

			/* Check for hashes */
			ret = ASN1_OCTET_STRING_cmp(tmp->nameHash, 
							b->issuerNameHash);
			if (ret) {
				// syslog( LOG_ERR, "%s:%d CA mismatch (%ld)",
				// 	__FILE__,__LINE__, ret);
				continue;
			}

			ret = ASN1_OCTET_STRING_cmp(tmp->keyHash,
							b->issuerKeyHash);
			if (ret) {
				// syslog( LOG_ERR, "%s:%d CA mismatch (%d)",
				// 	__FILE__,__LINE__, ret);
				continue;
			}

			/* If here we have found it! */
			if (!ret) return( conf->ca_list[i] );
		}
	}

	/* Here we have not found any suitable CA */
	return(NULL);
}

X509_REVOKED *ocspd_X509_REVOKED_find (CA_LIST_ENTRY *ca, ASN1_INTEGER *serial){

	X509_REVOKED *r = NULL;
	/*
	ASN1_ENUMERATED *asn = NULL;
	int crit, idx;
	*/

	int curr = 0;
	int start = 0;
	int cont = 1;
	int end, cmp_val;
	int found = 0;

	/* If no entries are in the list, return directly */
	if( !(ca) || !(ca->crl) || !(ca->crl_list)) return (r);
 
	/* Set the end point to the last one */
	end = sk_X509_REVOKED_num(ca->crl_list) - 1;
	if( end < 0 ) return (r);

	while( cont == 1 ) {
		/* We have not found the entry */
		if( end < start ) break;

		/* Calculate the middle between start and end */
		curr = (int) ((end - start) / 2) + start;

		/* Get the entry from the stack */
		r = sk_X509_REVOKED_value(ca->crl_list, curr);

		/* Compare the two serials */
		cmp_val = ASN1_INTEGER_cmp(r->serialNumber, serial);

		if( cmp_val > 0 ) {
			end = curr - 1;
			continue;
		} else if ( cmp_val < 0 ) {
			start = curr + 1;
			continue;
		} else {
			/* Entry Found ! */
			cont = 0;
			found = 1;
			break;
		}
	}
	if( found )
		return (r);
	else
		return(NULL);

}

